#include<thrust/device_vector.h>
#include<cuda_runtime_api.h>
#include<cuda_runtime.h>
#include<stdio.h>
#include<thrust/execution_policy.h>
#include <cub/cub.cuh>
#include <thrust/copy.h>

__device__ int MAX_RECURSION = 24;
__device__ int WARP_SIZE = 32;

__global__ void vectorPrinter(int *vector, int init, int nElements)
{
    for(int i = init; i < nElements; i++)
        printf("%d%s", vector[i], ((i > 0 && ((i % 20) == 0)) ? "\n" : " "));
    printf("\n");
}

__device__ void swapKernel(int *firstKey, int *secondKey)
{
    int key = *firstKey;
    *firstKey = *secondKey;
    *secondKey = key;
}

__device__ void oddEvenSortKernel(int *vector, long init, long nElements)
{
    int target = 0;
    __shared__ int swaped;

    do
    {
        swaped = 0;

        for(int index = (init + target); index < (nElements - target); index += 2)
            if(vector[index] > vector[index + 1])
                swapKernel(&vector[index], &vector[index + 1]), swaped += 1;

        target = (target > 0) ? 0 : 1;
    } while (__syncthreads_count(swaped) != 0 ||  __syncthreads_count(target) != 0);
}

__device__ void pivotMedianOf3(int vector[], long firstIndex, long lastIndex, long *pivotIndex)
{    
    long index = (firstIndex + lastIndex) / 2;
    if(vector[firstIndex] > vector[index])
    {   if(vector[firstIndex] < vector[lastIndex])            
            *pivotIndex = firstIndex;
        else        
        {            
            if(vector[index] < vector[lastIndex])                
                *pivotIndex = lastIndex;            
            else                
                *pivotIndex = index;                    
        }    
    }else{        
        if(vector[index] < vector[lastIndex])            
            *pivotIndex = index;        
        else        
        {            
            if(vector[firstIndex] < vector[lastIndex])                
                *pivotIndex = lastIndex;            
            else                
                *pivotIndex = firstIndex;                    
        }    
    }
}

__global__ void quickSortKernel(int *vector, long firstIndex, long lastIndex)
{
    long pivotIndex;
    long firstTempIndex, lastTempIndex;
    int pivot;
    cudaStream_t leftStream, rightStream;

    if(((lastIndex - firstIndex) <= WARP_SIZE)){
        oddEvenSortKernel(vector, firstIndex, lastIndex+1);
        return;
    }
    else{
        pivotMedianOf3(vector, firstIndex, lastIndex, &pivotIndex);
        swapKernel(&vector[pivotIndex], &vector[lastIndex]);
        firstTempIndex = firstIndex;
        lastTempIndex = lastIndex - 1;
        pivot = vector[lastIndex];

        while(lastTempIndex >= firstTempIndex)
        {
            while(vector[firstTempIndex] < pivot)
                firstTempIndex++;
            
            while(vector[lastTempIndex] > pivot)
                lastTempIndex--;
            
            if(lastTempIndex >= firstTempIndex)
            {
                swapKernel(&vector[firstTempIndex], &vector[lastTempIndex]);
                firstTempIndex++;
                lastTempIndex--;
            }

        }
        swapKernel(&vector[firstTempIndex], &vector[lastIndex]);

        cudaStreamCreateWithFlags(&leftStream, cudaStreamDefault);
        quickSortKernel<<<1, 1, 0, leftStream>>>(vector, firstIndex, firstTempIndex-1);
        cudaStreamDestroy(leftStream);

        cudaStreamCreateWithFlags(&rightStream, cudaStreamDefault);
        quickSortKernel<<<1, 1, 0, rightStream>>>(vector, firstTempIndex, lastIndex);
        cudaStreamDestroy(rightStream);
    }
}

__global__ void intercalate(int *vector, int *tempVector, long firstIndex, long middle, long lastIndex)
{
    long leftPartIndex = firstIndex;
    long rightPartIndex = middle;
    long index = firstIndex;

    while((leftPartIndex < middle) && (rightPartIndex <= lastIndex))
    {
        if(vector[leftPartIndex] < vector[rightPartIndex])
        tempVector[index] = vector[leftPartIndex], leftPartIndex += 1;
        else
        tempVector[index] = vector[rightPartIndex], rightPartIndex += 1;
        index += 1;
    }

    while(leftPartIndex < middle)
        tempVector[index] = vector[leftPartIndex], leftPartIndex += 1, index += 1;

    while(rightPartIndex <= lastIndex)
        tempVector[index] = vector[rightPartIndex], rightPartIndex += 1, index += 1;

    cudaDeviceSynchronize(); 

    for(int i = firstIndex; i < lastIndex+1; i++)
        vector[i] = tempVector[i];

}

 
__global__ void mergeSortKernel(int *vector, int *tempVector, long firstIndex, long lastIndex)
{
    int middle;
    cudaStream_t leftStream, rightStream;
    
    if((lastIndex - firstIndex) > WARP_SIZE)
    {
        middle = (firstIndex + lastIndex) / 2;

        cudaStreamCreateWithFlags(&leftStream, cudaStreamDefault);
        mergeSortKernel<<<1, 1>>>(vector, tempVector, firstIndex, middle);
        
        cudaStreamCreateWithFlags(&rightStream, cudaStreamDefault);
        mergeSortKernel<<<1, 1>>>(vector, tempVector, middle + 1, lastIndex);
        cudaDeviceSynchronize();
        intercalate<<<1, 1>>>(vector, tempVector, firstIndex, middle + 1, lastIndex);
        cudaStreamDestroy(leftStream);
        cudaStreamDestroy(rightStream);
    }
    else{
        oddEvenSortKernel(vector, firstIndex, lastIndex+1);
    }
}

__global__ void indexSearchKernel(int * vector, int limit, int number, int * index)
{
    int threadId = (blockDim.x * blockIdx.x) + threadIdx.x;
    if(threadId < limit){
        if(vector[threadId] <= number && number < vector[threadId + 1])
            index[0] = threadId + 1;
    }
}

__global__ void pushElementsKernel(int * vector, int limit, int index)
{
    int threadId = (blockDim.x * blockIdx.x) + threadIdx.x;
    if(threadId >= index && threadId < limit){
        vector[threadId + 1] = vector[threadId];  
    }
}

__global__ void insertionSortKernel(int * vector, long nElements, int *index)
{
    __shared__ int number;
    int nThreadsPerBlock = 1024;
    int nBlocks = (nElements / nThreadsPerBlock) + 1;

    for(int i = 2; i < nElements+1; i++){
        
        index[0] = i;
        number = vector[i];
    
        indexSearchKernel<<<nBlocks, nThreadsPerBlock>>>(vector, i, number, index);
        cudaDeviceSynchronize();
        pushElementsKernel<<<nBlocks, nThreadsPerBlock>>>(vector, i, index[0]);
        cudaDeviceSynchronize();
        vector[index[0]] = number;
        cudaDeviceSynchronize();
    }
}

/**
*=================================================================================
* MergeSort, uma abordagem de baixo nível
* com paralelismo controlado
*=================================================================================
**/
__global__ void rankCalculatorKernel(int *a, int *b, int *rank, int na, int nb)
{
    int threadId = (blockDim.x * blockIdx.x) + threadIdx.x;
    int index = 0;
    
    if(threadId < na)
        while((index < nb) && (b[index] <= a[threadId]))
        {
            rank[threadId] += 1;
            index += 1;
        }
}

__global__ void rankSumCalculatorKernel(int *rankProper, int *rankCompared, int nElements)
{
    int threadId = (blockDim.x * blockIdx.x) + threadIdx.x;
    
    if(threadId < nElements)
        rankProper[threadId] += rankCompared[threadId];
}

__global__ void mergeSortKernel(int *a, int *b, int *c,  int *ra, int *rb, int na, int nb)
{
    int threadId = (blockDim.x * blockIdx.x) + threadIdx.x;
    
    if(threadId < na)
        c[ra[threadId] - 1] = a[threadId];
    cudaDeviceSynchronize();

    if(threadId < nb)
        c[rb[threadId] - 1] = b[threadId];
    cudaDeviceSynchronize();
}

void mergeSort(int *a, int *b, int *c, int na, int nb)
{
    int nThreadPerBlock = 1024;
    int nABlocks = na / nThreadPerBlock;
    int nBBlocks = nb / nThreadPerBlock;
    int nCBlocks = (na + nb) / nThreadPerBlock;

    thrust::device_vector<int> rankAA(na);
    thrust::device_vector<int> rankAB(na);
    thrust::device_vector<int> rankBB(nb);
    thrust::device_vector<int> rankBA(nb);
    rankCalculatorKernel<<<nABlocks, nThreadPerBlock>>>(a, b, thrust::raw_pointer_cast(&rankAB[0]), na, nb);
    rankCalculatorKernel<<<nBBlocks, nThreadPerBlock>>>(b, a, thrust::raw_pointer_cast(&rankBA[0]), nb, na);
    rankCalculatorKernel<<<nABlocks, nThreadPerBlock>>>(a, a, thrust::raw_pointer_cast(&rankAA[0]), na, na);
    rankCalculatorKernel<<<nBBlocks, nThreadPerBlock>>>(b, b, thrust::raw_pointer_cast(&rankBB[0]), nb, nb);
    //cudaDeviceSynchronize();
    
    rankSumCalculatorKernel<<<nABlocks, nThreadPerBlock>>>(thrust::raw_pointer_cast(&rankAA[0]), 
                                                            thrust::raw_pointer_cast(&rankAB[0]), na);
    //cudaDeviceSynchronize();
    rankSumCalculatorKernel<<<nBBlocks, nThreadPerBlock>>>(thrust::raw_pointer_cast(&rankBB[0]), 
                                                            thrust::raw_pointer_cast(&rankBA[0]), nb);
    //cudaDeviceSynchronize();

    mergeSortKernel<<<nCBlocks, nThreadPerBlock>>>(a, b, c, thrust::raw_pointer_cast(&rankAA[0]), 
                                                    thrust::raw_pointer_cast(&rankBB[0]), na, nb);

}

void thrustSort(thrust::device_vector<int>deviceVector, float *executedTime)
{
    cudaEvent_t startTime, endTime;
    
    cudaEventCreate(&startTime);
    cudaEventCreate(&endTime);
    
    cudaEventRecord(startTime, 0);

    thrust::sort(deviceVector.begin(), deviceVector.end());

    cudaEventRecord(endTime, 0);

    cudaEventSynchronize(endTime);

    cudaEventElapsedTime(executedTime, startTime, endTime);

}